// crc32_new is a re-implementation of the crc32 hash table from the 2021 edition
// fo the TDT4205 Compiler Construction course at Norwegian University of Science
// and Technology. This implementation should be well optimised for the Go language.

package src

import (
	"errors"
)

// ----------------------------
// ----- Type definitions -----
// ----------------------------

// hElement is an arbitrary hash table element with a string key.
type hElement struct {
	key  string      // The lookup key for this element.
	val  interface{} // The arbitrary value held by this element.
	next *hElement   // Pointer to next hElement that was hashed to same index, and is part of linked list.
}

// hTable is a hash table that uses the CRC32 algorithm for hashing.
type hTable struct {
	size     int        // Number of elements in hash table. Can exceed Cap as hElement support chaining (single linked list).
	elements []hElement // Underlying slice of hElements.
}

// ---------------------
// ----- Constants -----
// ---------------------

// -------------------
// ----- Globals -----
// -------------------

// ---------------------
// ----- Functions -----
// ---------------------

// newHashTable creates a new crc32 hash table that has an underlying slice of size n.
func newHashTable(n int) (hTable, error) {
	res := hTable{size: 0}
	if n > 0 {
		res.elements = make([]hElement, n)
	} else {
		return res, errors.New("cannot create hash table with negative or zero indices")
	}
	return res, nil
}

// Cap returns the capacity of the underlying slice that supports the hash table.
func (ht *hTable) Cap() int {
	return cap(ht.elements)
}

// Size returns the number of elements currently store in the hash table.
func (ht *hTable) Size() int {
	return ht.size
}

// load returns the utilisation (u) of the underlying slice that supports the hash table divided by the
// size of the hash table. This effectively tells how evenly distributed all the entries are. When the size
// of the hash table is larger than the capacity of the underlying slice the load value becomes larger
// than 1.0. For size lower or equal to the capacity the ideal load should be less than or equal to 1.0.
// If size == capacity and load == 1.0 there are no collisions and the hash table is perfectly distributed.
func (ht *hTable) load() float32 {
	return float32(ht.size) / float32(len(ht.elements))
}

// values returns a slice of all the values in the hTable. An hElement contains both the key and value,
// which effectively makes the returned slice a slice of key-value pairs. Slice is not sorted by key value.
func (ht *hTable) values() []hElement {
	res := make([]hElement, ht.size)
	idx := 0
	for _, e1 := range ht.elements {
		if e1.val != nil {
			res[idx] = e1
			idx++
			for e2 := e1.next; e2 != nil; e2 = e2.next {
				res[idx] = *e2
				idx++
			}
		}
	}
	return res
}

// insert takes a key value pair and inserts a new hElement into the hTable ht.
func (ht *hTable) insert(key string, val interface{}) {
	hash := crc32(key, len(key))
	idx := int(hash) % len(ht.elements)
	e := ht.elements[idx]

	if e.val == nil {
		// No entry at this position.
		ht.elements[idx] = hElement{
			key:  key,
			val:  val,
			next: e.next,
		}
		ht.size++
	} else if e.key == key {
		// Replace hElement at index idx.
		ht.elements[idx] = hElement{
			key:  key,
			val:  val,
			next: e.next,
		}
		ht.size++
	} else {
		// Hash collision: append linked list.
		for e1 := &(ht.elements[idx]); ; {
			if e1.next == nil {
				te := hElement{
					key:  key,
					val:  val,
					next: nil,
				}
				e1.next = &te
				ht.size++
				return
			} else {
				e1 = e1.next
			}
		}
	}
}

// lookup takes a key and looks in the hash table for a matching key-value pair.
// If the entry exist it is returned and bool is true, else the value returned is nil and false.
func (ht *hTable) lookup(key string) (interface{}, bool) {
	hash := crc32(key, len(key))
	idx := int(hash) % len(ht.elements)
	e := ht.elements[idx]

	// Check entry in table and start of linked list.
	if e.key == key {
		return e.val, true
	}

	// Iterate linked list for entry.
	for e1 := e.next; e1 != nil; e1 = e1.next {
		if e1.key == key {
			return e1.val, true
		}
	}

	// Entry doesn't exist.
	return nil, false
}

// remove takes a key and removes the hElement with matching key if it exists.
// Returns true if an hElement was deleted. Return false if no hElement was deleted.
func (ht *hTable) remove(key string) bool {
	hash := crc32(key, len(key))
	idx := int(hash) % len(ht.elements)

	// Check entry in table and start of linked list.
	if ht.elements[idx].key == key {
		if ht.elements[idx].next == nil {
			ht.elements[idx] = hElement{}
		} else {
			ht.elements[idx] = *(ht.elements[idx].next)
		}
		ht.size--
		return true
	}

	// Iterate linked list for entry.
	prev := &(ht.elements[idx])
	for e1 := prev.next; e1 != nil; e1 = e1.next {
		if e1.key == key {
			// Remove this element.
			prev.next = e1.next
			return true
		}
		prev = e1
	}

	// Entry doesn't exist.
	return false
}
