// crc32_opt defines a hash table that only accepts string as key and Symbol as data values.
// It uses a bucket table with 8 pair buckets, just like the Go hash map.
//
// Buckets are appended during insertion and last appended bucket is full.
// Buckets are deallocated when elements are removed and linked bucket is empty and previous
// bucket is not full.
//
// When elements are removed the last inserted element is popped and put into the position of
// the removed element, if the removed element was not itself the last element.

package src

import (
	"errors"
)

// ----------------------------
// ----- Type definitions -----
// ----------------------------

// hAlg defines a hashing algorithm that takes a string key and returns an integer >= 0.
type hAlg func(string) int

// Symbol is a placeholder data element that represents an entry in the symbol table.
type Symbol struct {
	typ  int    // Symbol type.
	name string // Name of identifier.
}

// bElement holds a key value pair.
type bElement struct {
	key string // Key of element.
	val Symbol // Value of element.
	use bool   // Set true if element is stored, false if element doesn't hold data..
}

// bucket is a hash table bucket that holds multiple bElement.
type bucket struct {
	elements [bucketSize]bElement // 8 key-value paris.
	idx      int                  // Index of last inserted element.
	next     *bucket              // Linked list of buckets if bucket is full.
}

// bTable is a hash table with bucket tables.
type bTable struct {
	size    int      // Number of elements stored in hash table.
	buckets []bucket // Slice of buckets.
	alg     hAlg     // Hashing algorithm.
}

// ---------------------
// ----- Constants -----
// ---------------------

const (
	bucketSize = 8 // Same as Go hashmap.
)

// -------------------
// ----- Globals -----
// -------------------

// ---------------------
// ----- Functions -----
// ---------------------

// createHashTableOpt creates a new bTable and returns a pointer referencing the newly created hash table.
// If the input argument n is negative or zero, an error is returned. If the input argument alg is nil, the
// default algorithm, CRC32, is used for hashing lookup keys.
func newHashTableOpt(n int, alg hAlg) (*bTable, error) {
	if n < 1 {
		return nil, errors.New("cannot create empty hash table or hash table with negative size")
	}
	res := bTable{
		size:    0,
		buckets: make([]bucket, n),
		alg:     alg,
	}
	if alg == nil {
		res.alg = crc32s
	}
	return &res, nil
}

// Cap returns the size of the underlying slice supporting the table of buckets.
func (bt *bTable) Cap() int {
	return len(bt.buckets)
}

// Size returns the number of elements stored in the hash table.
func (bt *bTable) Size() int {
	return bt.size
}

// Load returns the load factor of the hash table. This effectively tells how evenly distributed all the entries are.
// When the size of the hash table is larger than the capacity of the underlying slice the load value becomes larger
// than 1.0. For size lower or equal to the capacity the ideal load should be less than or equal to 1.0.
// If size == capacity and load == 1.0 there are no collisions and the hash table is perfectly distributed.
func (bt *bTable) Load() float32 {
	return float32(bt.size) / float32(len(bt.buckets))
}

// Distro returns a fraction of how many buckets are used out of the total bucket count.
func (bt *bTable) Distro() float32 {
	m := 0
	for _, e1 := range bt.buckets {
		if e1.idx > 0 {
			m++
		}
	}
	return float32(m) / float32(len(bt.buckets))
}

// Insert inserts, and possibly overwrites, a Symbol in(to) the hash table.
// One can always expect this function to succeed.
func (bt *bTable) Insert(key string, data Symbol) {
	hash := bt.alg(key)
	idx := hash % len(bt.buckets)

	// Check for existing entry.
	for e1 := &(bt.buckets[idx]); e1 != nil; e1 = e1.next {
		for _, e2 := range e1.elements {
			if e2.key == key {
				e2.val = data
				return
			}
		}
	}

	// No entries found, append to bucket list.
	for e1 := &(bt.buckets[idx]); ; {
		// Find last appended bucket.
		if e1.next != nil {
			e1 = e1.next
		} else {
			// Append to this bucket.
			if e1.idx == bucketSize {
				// Bucket is full; append new bucket.
				e1.next = &bucket{
					elements: [bucketSize]bElement{},
					idx:      0,
					next:     nil,
				}
				e1 = e1.next
			}

			// Insert entry.
			e1.elements[e1.idx] = bElement{
				key: key,
				val: data,
				use: true,
			}
			e1.idx++
			bt.size++
			return
		}
	}
}

// Lookup returns the Symbol given input key if it exists. Returns true if Symbol was found, else it returns false.
func (bt *bTable) Lookup(key string) (Symbol, bool) {
	hash := bt.alg(key)
	idx := hash % len(bt.buckets)

	for e1 := &(bt.buckets[idx]); e1 != nil; e1 = e1.next {
		// Check all buckets for entry.
		for i2 := 0; i2 < e1.idx; i2++ {
			if e1.elements[i2].key == key {
				// Found entry with correct key.
				return e1.elements[i2].val, true
			}
		}
	}

	return Symbol{}, false
}

// Remove deletes a Symbol, given key, from the hash table if it exists.
// Returns true if Symbol was removed, else it returns false.
func (bt *bTable) Remove(key string) bool {
	hash := bt.alg(key)
	idx := hash % len(bt.buckets)

	var prev *bucket // Used for tracking previous element in linked list.
	for e1 := &(bt.buckets[idx]); e1 != nil; e1 = e1.next {
		for i2 := 0; i2 < e1.idx; i2++ {
			if e1.elements[i2].key == key {
				// Delete this element.
				if e1.idx == bucketSize && e1.next != nil {
					// Pop element from bucket list.
					if pop, err := e1.next.pop(); err != nil {
						// No elements in succeeding bucket(s).
						e1.elements[i2] = bElement{
							key: "",
							val: Symbol{},
							use: false,
						}
						e1.idx--
					} else {
						e1.elements[i2] = pop
					}
				} else {
					// Pop element from this bucket.
					if e1.idx > i2+1 {
						e1.elements[i2] = e1.elements[e1.idx-1]
						e1.elements[e1.idx] = bElement{}
					} else {
						e1.elements[i2] = bElement{}
					}
					e1.idx--
				}

				// Remove trailing bucket if second-to-last bucket is not full.
				if e1.idx == 0 && prev != nil && prev.idx < bucketSize {
					// Remove this bucket.
					prev.next = nil
				}
				bt.size--
				return true
			}
		}
		prev = e1
	}

	return false
}

// shift moves all trailing Symbols throughout the bucket list one index closer to the origin.
func (b *bucket) pop() (bElement, error) {
	var prev *bucket
	e := b
	for ; e != nil; e = e.next {
		// Find last appended bucket.
		prev = e
	}

	if e.idx == 0 {
		// Empty bucket found, check previous.
		if prev == nil {
			// Cannot pop from previous, because there is nothing to pop.
			return bElement{}, errors.New("no element to pop")
		}
		s := prev.elements[bucketSize-1]
		prev.elements[bucketSize-1] = bElement{}
		prev.idx--
		prev.next = nil // Remove linked bucket, because this is now not full.
		return s, nil
	}

	// Remove and return last inserted element of this bucket.
	s := e.elements[e.idx]
	e.elements[e.idx] = bElement{}
	e.idx--
	return s, nil
}

// crc32s takes a string input and generates a CRC32 hash defined by the IEEE 802.3 CRC32 algorithm.
func crc32s(s string) int {
	data := []uint8(s)
	hash := uint32(0xFFFFFFFF)
	for i1 := 0; i1 < len(s); i1++ {
		hash = (hash >> 8) ^ crc_table[data[i1]^uint8(hash)]
	}
	return int(hash ^ 0xFFFFFFFF)
}

// djb2 takes a string input and generates a DJB2 hash defined by Dan Bernstein.
func djb2(s string) int {
	hash := 5381
	for _, e1 := range s {
		hash = ((hash << 5) + hash) + int(e1)
	}
	return hash
}

