package src

import (
	"fmt"
	"testing"
)

// ----------------------------
// ----- Type definitions -----
// ----------------------------

// ----------------------
// ----- Constants ------
// ----------------------

const i = 10 // Number of buckets and/or size of underlying slice.

// --------------------
// ----- Globals ------
// --------------------

// Test data.
var symbols = [...]Symbol{
	{typ: 1, name: "varA"},
	{typ: 1, name: "varB"},
	{typ: 2, name: "varC"},
	{typ: 2, name: "varD"},
	{typ: 3, name: "varE"},
	{typ: 3, name: "varF"},
	{typ: 6, name: "varG"},
	{typ: 7, name: "varH"},
	{typ: 8, name: "varI"},
	{typ: 1, name: "varJ"},
}

// ----------------------
// ----- Functions ------
// ----------------------

// String returns a print friendly string representation of the Symbol.
func (s *Symbol) String() string {
	return fmt.Sprintf("%d [%q]", s.typ, s.name)
}

// TestCRC32p tests and verifies the operation fo the ported tlhash crc32 hash table.
func TestCRC32p(t *testing.T) {
	// Init.
	ht := tlhash_t{}
	if ok := tlhash_init(&ht, i); ok != TLHASH_SUCCESS {
		t.Fatal("could not initialise crc32 hash table")
	}

	// Before insertion.
	size := ht.tlhash_size()
	if size != 0 {
		t.Errorf("expected hash table size 0, got: %d", size)
	}

	// Insert test data.
	for _, e1 := range symbols {
		if ok := tlhash_insert(&ht, e1.name, len(e1.name), e1); ok != TLHASH_SUCCESS {
			t.Errorf("expected successful insert, got %d", ok)
		}
	}

	// After insertion.
	size = ht.tlhash_size()
	if size != len(symbols) {
		t.Errorf("expected hash table size %d, got: %d", len(symbols), size)
	}
	data := make([]interface{}, size)
	tlhash_values(&ht, data)
	for _, e1 := range data {
		if e1 == nil {
			t.Errorf("expected element, got <nil>")
		}
	}

	// Element lookup.
	for _, e1 := range symbols {
		var e interface{}
		if ok := tlhash_lookup(&ht, e1.name, len(e1.name), &e); ok == TLHASH_SUCCESS {
			if el := e.(Symbol); el.name != e1.name {
				t.Errorf("expected to find entry with name %q, got %q", e1.name, el.name)
			}
		} else {
			t.Errorf("expected to find entry with key %q, got none", e1.name)
		}
	}

	// Element removal.
	for _, e1 := range symbols {
		if ok := tlhash_remove(&ht, e1.name, len(e1.name)); ok != TLHASH_SUCCESS {
			t.Errorf("expected to remove element with key %q, got no removal", e1.name)
		}
	}

	// Check that size of hash table is zero after removing all symbols.
	if n := ht.tlhash_size(); n != 0 {
		t.Errorf("expected 0 entries in hash table, got %d", n)
	}

	// Cleanup.
	if ok := tlhash_finalize(&ht); ok != TLHASH_SUCCESS {
		t.Fatal("could not finalise crc32 hash table")
	}
}

// TestCRC32re tests and verifies the operation of the re-implemented tlhash crc32 hash table.
func TestCRC32re(t *testing.T) {
	ht, err := newHashTable(i)
	if err != nil {
		t.Fatal(err)
	}

	// Test size equals 0, and Cap equals 10.
	if ht.size != 0 {
		t.Errorf("expected table size 0, got %d", ht.size)
	}

	if ht.Cap() != i {
		t.Errorf("expected table capacity %d, got %d", i, ht.Cap())
	}

	// Test insertion and size increase.
	ht.insert(symbols[0].name, symbols[0])
	if ht.size != 1 {
		t.Errorf("expected table size 1, got %d", ht.size)
	}

	// Test lookup.
	if _, ok := ht.lookup(symbols[0].name); !ok {
		t.Errorf("expected to find symbol with name %q, got no match", symbols[0].name)
	}

	// Test deletion and size decrease.
	if ok := ht.remove(symbols[0].name); !ok {
		t.Errorf("expected removal of entry, but no entry was deleted")
	}
	if ht.size != 0 {
		t.Errorf("expected table size 0, got %d", ht.size)
	}

	// Test deletion when entry doesn't exist.
	if ok := ht.remove(symbols[0].name); ok {
		t.Errorf("expected no deletion, got deletion")
	}

	// Test should not find entry.
	if e, ok := ht.lookup(symbols[0].name); ok {
		t.Errorf("expected to find no match, got symbol with name %q", e.(Symbol).name)
	}

	// Insert all symbols.
	for i1, e1 := range symbols {
		if ht.size != i1 {
			t.Errorf("expected table size %d, got %d", i1, ht.size)
		}
		ht.insert(e1.name, e1)
		if ht.size != i1+1 {
			t.Errorf("expected table size %d, got %d", i1+1, ht.size)
		}
	}

	// Verify length again.
	if ht.size != len(symbols) {
		t.Errorf("expected table size %d, got %d", len(symbols), ht.size)
	}

	// Verify that all symbols are in hash table.
	for _, e1 := range symbols {
		if e, ok := ht.lookup(e1.name); !ok {
			t.Errorf("expected to find element with key %s, got no match", e1.name)
		} else {
			he := e.(Symbol)
			if he.name != e1.name {
				t.Errorf("expected element with name %q, got %q", e1.name, he.name)
			}
			if he.typ != e1.typ {
				t.Errorf("expected element with typ %d, got %d", e1.typ, he.typ)
			}
		}
	}

	// Delete all items.
	for i1, e1 := range symbols {
		if ok := ht.remove(e1.name); !ok {
			t.Errorf("expected to delete element with key %q, got no deletion", e1.name)
		}
		if ht.size != len(symbols)-i1-1 {
			t.Errorf("expected hash table size to be %d, got %d", len(symbols)-i1-1, ht.size)
		}
	}

	if ht.size != 0 {
		t.Errorf("expected hash table size to be 0, got %d", ht.size)
	}
}

// TestCRC32opt tests and verifies the operation of the optimised re-implementation of the tlhash crc32 hash table.
func TestCRC32opt(t *testing.T) {
	ht, err := newHashTableOpt(i, crc32s)
	if err != nil {
		t.Fatal(err)
	}

	// Test size equals 0, and Cap equals 10.
	if ht.size != 0 {
		t.Errorf("expected table size 0, got %d", ht.size)
	}

	if ht.Cap() != i {
		t.Errorf("expected table capacity %d, got %d", i, ht.Cap())
	}

	// Test insertion and size increase.
	ht.Insert(symbols[0].name, symbols[0])
	if ht.size != 1 {
		t.Errorf("expected table size 1, got %d", ht.size)
	}

	// Test lookup.
	if _, ok := ht.Lookup(symbols[0].name); !ok {
		t.Errorf("expected to find symbol with name %q, got no match", symbols[0].name)
	}

	// Test deletion and size decrease.
	if ok := ht.Remove(symbols[0].name); !ok {
		t.Errorf("expected removal of entry, but no entry was deleted")
	}
	if ht.size != 0 {
		t.Errorf("expected table size 0, got %d", ht.size)
	}

	// Test deletion when entry doesn't exist.
	if ok := ht.Remove(symbols[0].name); ok {
		t.Errorf("expected no deletion, got deletion")
	}

	// Test should not find entry.
	if e, ok := ht.Lookup(symbols[0].name); ok {
		t.Errorf("expected to find no match, got symbol with name %q", e.name)
	}

	// Insert all symbols.
	for i1, e1 := range symbols {
		if ht.size != i1 {
			t.Errorf("expected table size %d, got %d", i1, ht.size)
		}
		ht.Insert(e1.name, e1)
		if ht.size != i1+1 {
			t.Errorf("expected table size %d, got %d", i1+1, ht.size)
		}
	}

	// Verify length again.
	if ht.size != len(symbols) {
		t.Errorf("expected table size %d, got %d", len(symbols), ht.size)
	}

	// Verify that all symbols are in hash table.
	for _, e1 := range symbols {
		if e, ok := ht.Lookup(e1.name); !ok {
			t.Errorf("expected to find element with key %s, got no match", e1.name)
		} else {
			if e.name != e1.name {
				t.Errorf("expected element with name %q, got %q", e1.name, e.name)
			}
			if e.typ != e1.typ {
				t.Errorf("expected element with typ %d, got %d", e1.typ, e.typ)
			}
		}
	}

	// Delete all items.
	for i1, e1 := range symbols {
		if ok := ht.Remove(e1.name); !ok {
			t.Errorf("expected to delete element with key %q, got no deletion", e1.name)
		}
		if ht.size != len(symbols)-i1-1 {
			t.Errorf("expected hash table size to be %d, got %d", len(symbols)-i1-1, ht.size)
		}
	}

	if ht.size != 0 {
		t.Errorf("expected hash table size to be 0, got %d", ht.size)
	}
}

// TestDistro measures the distribution factor of both CRC32 and the Go hashmap hashig algorithms.
func TestDistro(t *testing.T) {
	ht := bTable{
		size:    0,
		buckets: make([]bucket, i),
		alg:     crc32s,
	}

	goh := make(map[string]Symbol, 10)

	for _, e1 := range symbols {
		ht.Insert(e1.name, e1)
		goh[e1.name] = e1
	}
	fmt.Printf("CRC32 distro: %f\n", ht.Distro())

	ht = bTable{
		size:    0,
		buckets: make([]bucket, i),
		alg:     djb2,
	}

	for _, e1 := range symbols {
		ht.Insert(e1.name, e1)
	}
	fmt.Printf("DJB2 distro: %f\n", ht.Distro())
}

// ------------------------
// ----- CRC32 Ported -----
// ------------------------

// BenchmarkCRC32pInsert benchmarks the Go port of the CRC32 hash table insert operation.
func BenchmarkCRC32pInsert(b *testing.B) {
	for n := 0; n < b.N; n++ {
		ht := tlhash_t{}
		tlhash_init(&ht, i)
		for _, e1 := range symbols {
			tlhash_insert(&ht, e1.name, len(e1.name), e1)
		}
		tlhash_finalize(&ht)
	}
}

// BenchmarkCRC32pLookup benchmarks the Go port of the CRC32 hash table lookup operation.
func BenchmarkCRC32pLookup(b *testing.B) {
	ht := initDataTlhash(i, b)
	for n := 0; n < b.N; n++ {
		for _, e1 := range symbols {
			var p interface{}
			tlhash_lookup(ht, e1.name, len(e1.name), &p)
		}
	}
	tlhash_finalize(ht)
}

// BenchmarkCRC32pRemove benchmarks the Go port of the CRC32 hash table remove operation.
func BenchmarkCRC32pRemove(b *testing.B) {
	for n := 0; n < b.N; n++ {
		ht := initDataTlhash(i, b)
		for _, e1 := range symbols {
			tlhash_remove(ht, e1.name, len(e1.name))
		}
		tlhash_finalize(ht)
	}
}

// --------------------------------
// ----- CRC32 Re-implemented -----
// ---------------------------------

// BenchmarkCRC32reInsert benchmarks the Go implementation of the CRC32 hash table insert.
func BenchmarkCRC32reInsert(b *testing.B) {
	for n := 0; n < b.N; n++ {
		ht := hTable{
			size:     0,
			elements: make([]hElement, i),
		}
		for _, e1 := range symbols {
			ht.insert(e1.name, e1)
		}
	}
}

// BenchmarkCRC32reLookup benchmarks the Go implementation of the CRC32 hash table lookup.
func BenchmarkCRC32reLookup(b *testing.B) {
	ht := initDataHT(i, b)
	for n := 0; n < b.N; n++ {
		for _, e2 := range symbols {
			ht.lookup(e2.name)
		}
	}
}

// BenchmarkCRC32reRemove benchmarks the Go implementation of the CRC32 hash table removal.
func BenchmarkCRC32reRemove(b *testing.B) {
	for n := 0; n < b.N; n++ {
		ht := initDataHT(i, b)
		for _, e1 := range symbols {
			ht.remove(e1.name)
		}
	}
}

// ----------------------
// ----- Go Hashmap -----
// ----------------------

// BenchmarkGohInsert benchmarks the Go built-in hash map insert operation.
func BenchmarkGohInsert(b *testing.B) {
	for n := 0; n < b.N; n++ {
		ht := make(map[string]Symbol, 10)
		for _, e1 := range symbols {
			ht[e1.name] = e1
		}
	}
}

// BenchmarkGohLookup benchmarks the Go built-in hash map lookup operation.
func BenchmarkGohLookup(b *testing.B) {
	ht := initDataHM(i, b)
	for n := 0; n < b.N; n++ {
		for _, e1 := range symbols {
			_ = (*ht)[e1.name]
		}
	}
}

// BenchmarkGohRemove benchmarks the Go built-in hash map remove operation.
func BenchmarkGohRemove(b *testing.B) {
	for n := 0; n < b.N; n++ {
		ht := initDataHM(i, b)
		for _, e1 := range symbols {
			delete(*ht, e1.name)
		}
	}
}

// ---------------------------
// ----- CRC32 Optimised -----
// ---------------------------

// BenchmarkCRC32optInsert benchmarks the optimised Go implementation of the CRC32 hash table insert.
func BenchmarkCRC32optInsert(b *testing.B) {
	for n := 0; n < b.N; n++ {
		ht := bTable{
			size:    0,
			buckets: make([]bucket, i),
			alg:     crc32s,
		}
		for _, e1 := range symbols {
			ht.Insert(e1.name, e1)
		}
	}
}

// BenchmarkCRC32optLookup benchmarks the optimised Go implementation of the CRC32 hash table lookup.
func BenchmarkCRC32optLookup(b *testing.B) {
	ht := initDataBT(i, crc32s, b)
	for n := 0; n < b.N; n++ {
		for _, e2 := range symbols {
			ht.Lookup(e2.name)
		}
	}
}

// BenchmarkCRC32optRemove benchmarks the optimised Go implementation of the CRC32 hash table removal.
func BenchmarkCRC32optRemove(b *testing.B) {
	for n := 0; n < b.N; n++ {
		ht := initDataBT(i, crc32s, b)
		for _, e1 := range symbols {
			ht.Remove(e1.name)
		}
	}
}

// --------------------------
// ----- DJB2 Optimised -----
// --------------------------

// BenchmarkDJB2optInsert benchmarks the Go implementation of the DJB2 hash table insert.
func BenchmarkDJB2optInsert(b *testing.B) {
	for n := 0; n < b.N; n++ {
		ht := bTable{
			size:    0,
			buckets: make([]bucket, i),
			alg:     djb2,
		}
		for _, e1 := range symbols {
			ht.Insert(e1.name, e1)
		}
	}
}

// BenchmarkDJB2optLookup benchmarks the Go implementation of the DJB2 hash table lookup.
func BenchmarkDJB2optLookup(b *testing.B) {
	ht := initDataBT(i, djb2, b)
	for n := 0; n < b.N; n++ {
		for _, e2 := range symbols {
			ht.Lookup(e2.name)
		}
	}
}

// BenchmarkDJB2optRemove benchmarks the Go implementation of the DJB2 hash table removal.
func BenchmarkDJB2optRemove(b *testing.B) {
	for n := 0; n < b.N; n++ {
		ht := initDataBT(i, djb2, b)
		for _, e1 := range symbols {
			ht.Remove(e1.name)
		}
	}
}

// -------------------
// ----- Helpers -----
// -------------------

// initDataHT initiates a hash table with all the symbols inserted.
func initDataHT(n int, b *testing.B) *hTable {
	b.Helper()
	res := hTable{
		size:     0,
		elements: make([]hElement, n),
	}
	for _, e1 := range symbols {
		res.insert(e1.name, e1)
	}
	return &res
}

// initDataBT initiates an optimised bucketed hash table with all the symbols inserted.
func initDataBT(n int, hf hAlg, b *testing.B) *bTable {
	b.Helper()
	res := bTable{
		size:    0,
		buckets: make([]bucket, n),
		alg:     crc32s,
	}
	for _, e1 := range symbols {
		res.Insert(e1.name, e1)
	}
	return &res
}

// initDataTlhash returns a Go port tlhash table with all symbols inserted.
func initDataTlhash(n int, b *testing.B) *tlhash_t {
	b.Helper()
	ht := tlhash_t{}
	tlhash_init(&ht, n)
	for _, e1 := range symbols {
		tlhash_insert(&ht, e1.name, len(e1.name), e1)
	}
	return &ht
}

// initDataHM initiates a built-in hash map with all the symbols inserted.
func initDataHM(n int, b *testing.B) *map[string]Symbol {
	b.Helper()
	res := make(map[string]Symbol, n)
	for _, e1 := range symbols {
		res[e1.name] = e1
	}
	return &res
}
